# import asyncio
import aiohttp_jinja2


@aiohttp_jinja2.template('home.j2.html')
def default(request):
    return {}


# @asyncio.coroutine
# def handler(request):
#     context = {'name': 'Andrew', 'surname': 'Svetlov'}
#     response = aiohttp_jinja2.render_template(
#         'tmpl.j2.html', request, context)
#     response.headers['Content-Language'] = 'ru'
#     return response



"""
txtulip: Twisted reactor based on Tulip/Asyncio/Trollius
========================================================

``txtulip`` is licensed under the MIT open source license, and maintained by
Itamar Turner-Trauring.

``txtulip`` can be found at https://github.com/itamarst/txtulip
"""



from scrapybox import api, views


def add(app):
    ar = app.router.add_route

    # API
    ar('POST', '/api/eval',         api.parse_post)
    ar('GET',  '/api/eval/{code}',  api.parse_get)
    ar('GET',  '/api/delay/{time}', api.delay)
    ar('GET',  '/api/hello',        api.hello)

    # Client
    ar('*', '/', views.default)

    # Static
    app.router.add_static('/static', app['static_path'])



import types
import asyncio
import logging
import aiohttp.web
import scrapy.signals
import scrapy.crawler

from scrapy.utils.project import get_project_settings

import scrapybox.crawler.spiders

SETTINGS = get_project_settings()

logger = logging.getLogger(__name__)


async def poll(crawler, time=1):
    while crawler.crawling:
        await asyncio.sleep(time)


async def parse_post(request):
    """
    $ curl localhost:8080/api/eval -d start_urls=[\"http://scrapinghub.com\",\"http://scrapy.org\"]
    Scraped 2 items: [{'response': <200 http://scrapinghub.com>}, {'response': <200 http://scrapy.org>}]

    <Request POST /api/eval > <MultiDictProxy('start_url': 'scrapy.org', 'parse': 'text = response.xpath(\'id("scrapy-logo")/following-sibling::p\')\r\nitem = dict(text=text.extract())\r\n', 'yield_item': 'on')>
    2016-03-25 12:08:12 [aiohttp.access] INFO: 127.0.0.1 - - [25/Mar/2016:19:08:12 +0000] "POST /api/eval HTTP/1.1" 200 188 "-" "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Ubuntu Chromium/48.0.2564.116 Chrome/48.0.2564.116 Safari/537.36"
    Scraped 1 items: [{'text': ['<p>An open source and collaborative framework for extracting the data you need from websites.\n      </p>', '<p>In a fast, simple, yet extensible way.</p>']}]

    http://stackoverflow.com/questions/35256152/can-eval-be-applied-on-a-compiled-ast-node-have-a-local-context
        self.emit ('\n{}\n', node.args [1] .s.format (* [
            eval (
                compile (
                    ast.Expression (arg),
                    '<string>',
                    'eval'
                ),
                {},
                {'include': include}
            )
            for arg in node.args [2:]
        ]))
    """
    await request.post()

    server_info = dict(
        request=dict(
            host=str(request.host),
            path=str(request.path),
            POST=dict(request.POST),
            query=str(request.query_string),
            method=str(request.method),
            scheme=str(request.scheme),
            cookies=dict(request.cookies),
            version=str(request.version),
            headers=dict(request.headers),
            content=list(request.content.read()),
            payload=list(request.payload.read()),
            has_body=bool(request.has_body),
        ),
    )

    class Spider(scrapybox.crawler.spiders.ExampleYieldSpider):
        pass

    if 'start_urls' in request.POST:
        value = request.POST['start_urls'].strip()
        if value:
            Spider.start_urls = eval(value)

    if 'start_url' in request.POST:
        value = request.POST['start_url'].strip()
        if value:
            url = scrapy.utils.url.add_http_if_no_scheme(value)
            Spider.start_urls = [url]

    if request.POST.get('parse'):
        def parse(self, response):
            exec(request.POST['parse'], globals(), locals())
            if request.POST.get('yield_item'):
                frame_item = locals()['item']
                yield frame_item
        Spider.parse = types.MethodType(parse, Spider)

    settings = dict(SETTINGS)

    settings['ROBOTSTXT_OBEY'] = 'settings.ROBOTSTXT_OBEY' in request.POST
    settings['HTTPCACHE_ENABLED'] = 'settings.HTTPCACHE_ENABLED' in request.POST
    if 'settings.USER_AGENT' in request.POST:
        settings['USER_AGENT'] = request.POST['settings.USER_AGENT']

    items = []

    def item_scraped(item, response, spider):
        items.append(item)

    crawler = scrapy.crawler.Crawler(Spider, settings)
    crawler.signals.connect(item_scraped, signal=scrapy.signals.item_scraped)
    logger.debug('API starting to crawl {}'.format(crawler))
    crawler.crawl()
    await poll(crawler)
    data = dict(items=items, scrapybox=server_info)

    return aiohttp.web.json_response(data)


async def parse_get(request):
    """
    $ curl "http://127.0.0.1:8080/api/parse/eval/response.status"
    200

    $ curl "http://127.0.0.1:8080/api/parse/eval/response.headers"
    {b'Last-Modified': [b'Fri, 09 Aug 2013 23:54:35 GMT'], b'Etag': [b'"359670651+gzip"'], b'X-Cache': [b'HIT'], b'Date': [b'Mon, 21 Mar 2016 20:17:19 GMT'], b'X-Ec-Custom-Error': [b'1'], b'Cache-Control': [b'max-age=604800'], b'Vary': [b'Accept-Encoding'], b'Content-Type': [b'text/html'], b'Server': [b'ECS (ftw/FBE4)'], b'Expires': [b'Mon, 28 Mar 2016 20:17:19 GMT']}

    $ curl "http://127.0.0.1:8080/api/parse/eval/response"
    <200 http://www.example.com/>

    $ curl "http://127.0.0.1:8080/api/parse/eval/self"
    <Spider 'parse' at 0x7f3a3af2fb38>

    $ curl "http://127.0.0.1:8080/api/parse/eval/self.__dict__"
    {'settings': <scrapy.settings.Settings object at 0x7f3a3af47e10>, 'result': {...}, 'crawler': <scrapy.crawler.Crawler object at 0x7f3a3af47b70>}
    """
    class Spider(scrapybox.crawler.spiders.ExampleSpider):
        result = None
        def parse(self, response):
            self.result = eval(self.code())
            yield

        @staticmethod
        def code():
            return request.match_info.get('code')

    crawler = scrapy.crawler.Crawler(Spider, SETTINGS)
    logger.debug('API starting to crawl {}'.format(crawler))
    crawler.crawl()
    await poll(crawler)
    text = '{}\n'.format(crawler.spider.result)

    return aiohttp.web.Response(body=text.encode('utf-8'))


async def delay(request):
    """
    $ curl localhost:8080/api/delay/15 &
    [1] 17172
    $ curl localhost:8080/api/delay/5 &
    [2] 17174
    $ curl localhost:8080/api/delay/1 &
    [3] 17177
    $ /api/delay/1
    /api/delay/5
    /api/delay/15
    """
    Spider = scrapybox.crawler.spiders.ExampleSpider
    crawler = scrapy.crawler.Crawler(Spider, SETTINGS)
    logger.debug('API starting to crawl {}'.format(crawler))
    crawler.crawl()
    await poll(crawler, float(request.match_info.get('time')))
    text = '{}\n'.format(request.path)

    return aiohttp.web.Response(body=text.encode('utf-8'))


def hello(request):
    """
    $ curl "http://127.0.0.1:8080/api/hello"
    Hello world!
    """
    text = 'Hello world!\n'
    return aiohttp.web.Response(body=text.encode('utf-8'))



"""
Scrapybox - A Scrapy server joint
"""
# __all__ = ['__version__', 'version_info', 'api', 'routes', 'server']

__version__ = '0.1'
version_info = tuple(int(v) if v.isdigit() else v
                     for v in __version__.split('.'))



import scrapybox.server._twisted_monkey_patches

import os
import asyncio
import logging
import aiohttp.web
import aiohttp_jinja2
# import aiohttp_debugtoolbar
import twisted.internet.reactor
import jinja2

from scrapy.utils.log import configure_logging
from scrapy.utils.project import get_project_settings

import scrapybox.server.routes

logger = logging.getLogger(__name__)
settings = get_project_settings()
user_path = os.path.join(os.getcwd(), 'scrapybox/user')


async def on_shutdown(app):
    """
    LOG_STDOUT

        If logging stdout then StreamLogger needs a flush() method.
        To recreate: enable LOG_STDOUT, run server, ^C interrupt:
        Exception ignored in: <scrapy.utils.log.StreamLogger object at 0x7f...>
        AttributeError: 'StreamLogger' object has no attribute 'flush'

    twisted_log.PythonLoggingObserver

        Observer is started in configure_logging, do we need to close it?

    Close all opened sockets manually:

        http://aiohttp.readthedocs.org/en/stable/web.html#graceful-shutdown

        for ws in app['websockets']:
            await ws.close(code=999, message='Server shutdown')
    """
    logger.info('Scrapybox server shutting down')
    # print(asyncio.Server.sockets, 'open sockets for asyncio')
    # asyncio.Server.close()
    # print(aiohttp.web.connections, 'open connections for aiohttp')
    app.shutdown()
    # aiohttp.web.finish_connections(5)
    app.cleanup()
    logger.info('Scrapybox server shutdown complete')


def main():
    """
    Scrapy pull request:
        configure_logging() should accept a config argument

    Note:
        scrapy.utils.log.TopLevelFormatter is cool
        need to access Scrapy loggger's handler and replace the filter with a new
        TopLevelFormatter with more names: e.g. ['scrapy', 'scrapybox', 'aiohttp']
    """
    configure_logging(settings)
    logger.info('Scrapybox server starting')
    # formatter = logging.Formatter(fmt=settings.get('LOG_FORMAT'),
    #                               datefmt=settings.get('LOG_DATEFORMAT'))
    # handler = logging.StreamHandler()
    # handler.setFormatter(formatter)
    # handler.setLevel(settings.get('LOG_LEVEL'))
    # logging.root.addHandler(handler)

    twisted.internet.reactor.run()
    logger.info('Twisted reactor running')

    app = aiohttp.web.Application(
        loop=asyncio.get_event_loop(),
        # middlewares=[aiohttp_debugtoolbar.toolbar_middleware_factory]
    )
    # aiohttp_debugtoolbar.setup(app)  # http://127.0.0.1:8080/_debugtoolbar
    aiohttp_jinja2.setup(app, loader=jinja2.FileSystemLoader(user_path))
    app.on_shutdown.append(on_shutdown)
    app['static_path'] = user_path

    scrapybox.server.routes.add(app)

    logger.info('Aiohttp server starting')
    aiohttp.web.run_app(app)

if __name__ == '__main__':
    main()



"""
Tulip-based reactor implementation.
"""
try:
    from asyncio import get_event_loop, new_event_loop
except ImportError:
    # Try importing trollius, the Python 2 backport of asyncio, if asyncio
    # is not found.
    from trollius import get_event_loop, new_event_loop

from zope.interface import implementer

from twisted.internet.base import DelayedCall
from twisted.internet.posixbase import PosixReactorBase
from twisted.python.log import callWithLogger
from twisted.internet.interfaces import IReactorFDSet


class _DCHandle(object):
    """
    Wrapper for asyncio.Handle to be used by DelayedCall.
    """
    def __init__(self, handle):
        self.handle = handle

    def cancel(self):
        self.handle.cancel()


@implementer(IReactorFDSet)
class AsyncioSelectorReactor(PosixReactorBase):
    """
    Reactor running on top of an asyncio SelectorEventLoop.
    """
    _asyncClosed = False

    def __init__(self, eventloop):
        self._asyncioEventloop = eventloop
        self._writers = set()
        self._readers = set()
        self._delayedCalls = set()
        PosixReactorBase.__init__(self)

    def _read_or_write(self, selectable, read):
        method = selectable.doRead if read else selectable.doWrite
        try:
            why = method()
        except Exception as e:
            why = e
        if why:
            self._disconnectSelectable(selectable, why, read)

    def addReader(self, reader):
        self._readers.add(reader)
        fd = reader.fileno()
        self._asyncioEventloop.add_reader(fd, callWithLogger, reader,
                                          self._read_or_write, reader, True)

    def addWriter(self, writer):
        self._writers.add(writer)
        fd = writer.fileno()
        self._asyncioEventloop.add_writer(fd, callWithLogger, writer,
                                          self._read_or_write, writer, False)

    def removeReader(self, reader):
        try:
            self._readers.remove(reader)
        except KeyError:
            pass
        fd = reader.fileno()
        if fd == -1:
            return
        self._asyncioEventloop.remove_reader(fd)

    def removeWriter(self, writer):
        try:
            self._writers.remove(writer)
        except KeyError:
            pass
        fd = writer.fileno()
        if fd == -1:
            return
        self._asyncioEventloop.remove_writer(fd)

    def removeAll(self):
        return self._removeAll(self._readers, self._writers)

    def getReaders(self):
        return list(self._readers)

    def getWriters(self):
        return list(self._writers)

    def getDelayedCalls(self):
        return list(self._delayedCalls)

    def iterate(self, timeout):
        self._asyncioEventloop.call_later(timeout + 0.01,
                                          self._asyncioEventloop.stop)
        self._asyncioEventloop.run_forever()

    def run(self, installSignalHandlers=True):
        self.startRunning(installSignalHandlers=installSignalHandlers)
        # self._asyncioEventloop.run_forever()  #SMA
        if self._justStopped:
            self._justStopped = False

    def stop(self):
        PosixReactorBase.stop(self)
        self.callLater(0, self.fireSystemEvent, "shutdown")

    def crash(self):
        PosixReactorBase.crash(self)
        self._asyncioEventloop.stop()

    def seconds(self):
        return self._asyncioEventloop.time()

    def callLater(self, seconds, f, *args, **kwargs):
        def run():
            dc.called = True
            self._delayedCalls.remove(dc)
            f(*args, **kwargs)
        handle = self._asyncioEventloop.call_later(seconds, run)
        dchandle = _DCHandle(handle)
        def cancel(dc):
            self._delayedCalls.remove(dc)
            dchandle.cancel()
        def reset(dc):
            dchandle.handle = self._asyncioEventloop.call_at(dc.time, run)

        dc = DelayedCall(self.seconds() + seconds, run, (), {},
                         cancel, reset, seconds=self.seconds)
        self._delayedCalls.add(dc)
        return dc

    def callWhenRunning(self, f, *args, **kwargs):
        g = lambda: f(*args, **kwargs)
        self._asyncioEventloop.call_soon_threadsafe(g)
    callFromThread = callWhenRunning


# Install some testing infrastructure; please don't look at this:
@staticmethod
def _reactorForTesting():
    loop = new_event_loop()
    loop.set_debug(True)
    return AsyncioSelectorReactor(loop)
def _installTestInfrastructure():
    from twisted.internet.test.reactormixins import ReactorBuilder
    ReactorBuilder._reactors.append("txtulip.reactor._reactorForTesting")
_installTestInfrastructure()


def install(eventloop=None):
    """
    Install a tulip-based reactor.
    """
    if eventloop is None:
        eventloop = get_event_loop()
    reactor = AsyncioSelectorReactor(eventloop)
    from twisted.internet.main import installReactor
    installReactor(reactor)



"""
Scrapybox crawler settings
"""
BOT_NAME = 'Scrapybox'

SPIDER_MODULES = ['scrapybox.crawler.spiders']
NEWSPIDER_MODULE = 'scrapybox.crawler.spiders'

USER_AGENT = 'Mozilla/4.0 Scrapybox (Linux x86_64) Scrapy/1.2.0dev2 Python/3.5.0+'

ROBOTSTXT_OBEY = True

CONCURRENT_REQUESTS = 1

HTTPCACHE_ENABLED = True

LOG_STDOUT = True

# DOWNLOAD_DELAY = 3
# COOKIES_ENABLED = False



import scrapy


class ExampleSpider(scrapy.Spider):
    name = 'example'
    start_urls = ('http://www.example.com/',)

    def parse(self, response):
        yield


class ExampleYieldSpider(ExampleSpider):
    name = 'example-yield'

    def parse(self, response):
        yield dict(
            request=dict(
                url=response.request.url,
                meta=response.request.meta,
                method=response.request.method,
                cookies=response.request.cookies,
                encoding=response.request.encoding,
                headers={
                    k.decode('utf-8'): [v.decode('utf-8') for v in l]
                    for k, l in response.request.headers.items()
                },
            ),
            response=dict(
                url=response.url,
                meta=response.meta,
                flags=response.flags,
                status=response.status,
                headers={
                    k.decode('utf-8'): [v.decode('utf-8') for v in l]
                    for k, l in response.headers.items()
                },
                body=dict(
                    head=response.css('head').extract_first()[:1000],
                    title=response.xpath('/html/head/title/text()').extract_first(),
                ),
            )
        )



import setuptools

setuptools.setup(
    name='scrapybox',
    version='0.1',
    license='BSD',
    url='https://github.com/stav/scrapybox',
    download_url='https://github.com/stav/scrapybox/archive/master.zip',
    description='A Scrapy GUI',
    long_description=open('README').read(),
    author='Steven Almeroth',
    author_email='sroth77@gmail.com',
    maintainer='Steven Almeroth',
    maintainer_email='sroth77@gmail.com',
    install_requires=['aiohttp', 'aiohttp-jinja2', 'cchardet'],
    packages=setuptools.find_packages(exclude=('tests', 'docs')),
    package_data={"": ["LICENSE", "README"]},
    include_package_data=True,
    keywords=[
        'scrapy',
        'async',
        'server',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Framework :: Scrapy',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Browsers',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    entry_points={
        'console_scripts': ['sbserve = scrapybox.server.server:main']
    },
)
